/** \file investment.cpp
	\brief Implementation of the investment model, using FlopC++
	\author Michal Kaut

	This version uses objects for models at different stages of the
	scenario tree. It is based on the \c stampl.cpp example from \a FlopC++
	distribution.

	Note that the code is meant as an illustrative example that mixes different
	styles to show more ways of doing things, something you most likely do \e not
	want to do in a real code. In addition, in a real code one would probably
	make the members private and write get/set functions where needed. This has
	been omitted here to make the example shorter.
*/

#include <iomanip>
#include "flopc.hpp"

// Change these two lines to use a different solver
#include "OsiClpSolverInterface.hpp"
#define OSI_SOLVER_INTERFACE OsiClpSolverInterface

using namespace std;
using namespace flopc;


/// This is the base class for describing a model in one node of the tree.
/** Note that variable \a x, as well as the vector of \a children, should not
    really be in this class, as they are not needed in all the nodes (they
    do not exist in leaves). It would thus be better to have an additional
    derived class for non-leaf nodes and then derive RootNode and MidStageNode
    from this one. This solution, however, brings some other problems... */
class StageNode {
	public:
		MP_set ASSETS;                ///< set of assets
		MP_index a;                   ///< index used in formulas
		MP_variable x;                ///< the "buy" variable, defined on ASSETS
		MP_expression objFunction;    ///< objective function at this node

		StageNode *ptParent;          ///< pointer to parent node
		vector<StageNode *> children; ///< pointers to children of this node
		int nodeNmb;                  ///< node number - used only for reporting
		double prob;                  ///< probability (unconditional) of the node

		StageNode(StageNode *ptPred, const int nmbAssets, const int nodeN,
		          const double condProb)
		: ASSETS(nmbAssets), x(ASSETS), ptParent(ptPred), nodeNmb(nodeN),
		  prob(condProb)
		{
			if (ptParent != NULL) {
				prob *= ptParent->prob;             // Compute the total probability
				ptParent->children.push_back(this); // Register with the parent
			}
		}
		virtual ~StageNode(){}

	protected:
		/// Create the objective function expression, recursively for all children
		/** This function is protected, as it only makes sense to call it in
		    the root, to create the complete objective function. **/
		virtual void make_obj_function_() {
			assert (children.size() > 0 && "Leaves should never call this function");
			// Create the objective recursively for all descendants of the node
			for (int i=0; i < (int) children.size(); i++) {
				children[i]->make_obj_function_();
			}
			// No objective at non-leaf nodes, so the objective value at this node
			// and all its descendants is a sum of obj. values of the children.
			objFunction = children[0]->objFunction;
			for (int i=1; i < (int) children.size(); i++) {
				objFunction = objFunction + children[i]->objFunction;
			}
		}
};

/// Class for the root node of the tree
class RootNode : public StageNode {
	public:
		MP_constraint initialBudget; ///< initial budget constraint

		RootNode(const int nmbAssets, const double initWealth)
		: StageNode(NULL, nmbAssets, 0, 1.0)
		{
			initialBudget() = sum(ASSETS, x(ASSETS)) == initWealth;
		}

		/// This is the public interface to the protected \c make_obj_function_()
		/** This is to prevent the user to call \c make_obj_function_() for other
		    nodes than the root, which does not make sense. **/
		void make_objective_function() {
			make_obj_function_();
		}
};

/// Class for the middle-nodes, i.e. all the nodes between the root and leaves
class MidStageNode : public StageNode {
	public:
		MP_constraint cashFlowBalance; ///< cash-flow balance constraint
		MP_data Return;                ///< returns of the assets at this node

		/// Constructs a \c MidStageNode object
		/** Here we use a <em>shallow copy</em> for the \c MP_data \a Return,
		    i.e. the return values in the constraints will be linked to the array
		    \a retVect. This means that if the external array changes before we
		    build the OSI object (using the \c attach method), the constraints will
		    be changed as well - and if the external array is deallocated, the
		    program will crash on calling the \c attach method! **/
		MidStageNode(StageNode *ptPred, const int nodeN, const double condProb,
		             double *ptRetVect)
		: StageNode(ptPred, ptPred->ASSETS.size(), nodeN, condProb),
		  Return(ptRetVect, ASSETS)
		{
			// This shows the use of MP_index in a formula
			cashFlowBalance = sum(ASSETS(a), ptParent->x(a) * Return(a))
			                  == sum(ASSETS(a), x(a));
		}
		virtual ~MidStageNode(){}
};

/// This is the class for the leaves, i.e. the last-stage nodes
class LeafNode : public StageNode {
	public:
		MP_variable w;              ///< shortage variable
		MP_variable y;              ///< surplus variable
		MP_constraint finalBalance; ///< constraint for the final balance
		MP_data Return;             ///< returns of the assets at this node

		/// Constructs a \a LeafNode object
		/** In this case, we use a <em>deep copy</em> for the \c MP_data \a Return,
		    i.e. the return values in the constraints are copied from the \a retVect
		    array to the constraints. This means that the \a retVect array can be
		    safely changed or deleted afterwards. **/
		LeafNode(StageNode *ptPred, const int nodeN, const double condProb,
		         const double *ptRetVect, const double capTarget)
		: StageNode(ptPred, ptPred->ASSETS.size(), nodeN, condProb), Return(ASSETS)
		{
			Return.value(ptRetVect); // Copy values from retVect to Return
			// This shows a formula without using an additional MP_index
			finalBalance = sum(ASSETS, ptParent->x(ASSETS) * Return(ASSETS))
			               + w() - y() == capTarget;
		}

	protected:
		/// version of \a make_obj_function_() for the leaves - no recursion
		void make_obj_function_() {
			objFunction = prob * (1.3 * w() - 1.1 * y());
		}
};


// -------------------------------------------------------------------------- //

/// Class describing the scenario-tree structure
class ScenTreeStruct {
	public:
		int nmbNodes;   ///< nodes are 0..nmbNodes-1, where 0 is root
		int firstLeaf;  ///< nodes firstLeaf..nmbNodes-1 are leaves

		/// Constructs the object
		ScenTreeStruct(const int nNodes, const int firstL)
		: nmbNodes(nNodes), firstLeaf(firstL)
		{}
		virtual ~ScenTreeStruct(){}

		/// This function returns the parent of a given node (and 0 for the root)
		/** In a general case, this would be given by a table, for balanced trees
		    one can use a simple formula. The question is what to do with the root:
		    should the function return 0, -1, or throw an exception? **/
		virtual int get_parent(int n) const = 0;
};

/// Class for balanced binary trees
class BinTreeStruct : public ScenTreeStruct {
	public:
		/// Constructs the object - 2^T-1 nodes, first leaf is 2^(T-1)-1
		BinTreeStruct(const int nmbStages)
		: ScenTreeStruct(pow(2.0, nmbStages) - 1, pow(2.0, nmbStages-1) - 1) {}

		int get_parent(int n) const {
			return (n-1) / 2;    // This gives: get_parent(0) = 0
		}
};

// -------------------------------------------------------------------------- //

int main()
{
	// DATA - This would be normally read from some external file
	// binary scenario tree with 4 stages: 15 nodes, firstLeaf = 7
	BinTreeStruct scTree(4);

	// model parameters
	enum {stocks, bonds, nmbAssets}; // assets to invest to; sets nmbAssets = 2
	double InitBudget = 55;          // initial budget
	double CapTarget = 80;           // capital target

	// vector of returns at the 14 non-root nodes
	double retData[][nmbAssets] = {
		{1.25, 1.14},
		{1.06, 1.16},
		{1.21, 1.17},
		{1.07, 1.12},
		{1.15, 1.18},
		{1.06, 1.12},
		{1.26, 1.13},
		{1.07, 1.14},
		{1.25, 1.15},
		{1.06, 1.12},
		{1.05, 1.17},
		{1.06, 1.15},
		{1.05, 1.14},
		{1.06, 1.12}
	};

	// ------------------------------------------------------------------------ //
	// Initialize the object for the model
	MP_model &model = MP_model::getDefaultModel();
	model.setSolver(new OSI_SOLVER_INTERFACE);
	#ifdef NDEBUG
		model.silent();  // non-debug, i.e. release version -> less output
	#else
		model.verbose(); // debug version -> show more output
	#endif

	// Create the scenario tree
	vector<StageNode *> treeNodes;
	treeNodes.push_back(new RootNode(nmbAssets, InitBudget));
	for (int i = 1; i < scTree.nmbNodes; i++) {
		int pred = scTree.get_parent(i);
		if (i < scTree.firstLeaf) {
			// Adding a mid-stage node
			treeNodes.push_back(
				// Remember that this uses a shallow copy of retData,
				// i.e. that it links the constraints to the retData array!
				new MidStageNode(treeNodes[pred], i, 1.0 / 2, retData[i-1])
			);
		} else {
			// Adding a leaf
			treeNodes.push_back(
				// This constructor uses a deep copy, so it does not create a link.
				new LeafNode(treeNodes[pred], i, 1.0 / 2, retData[i-1], CapTarget)
			);
		}
	}

	#ifndef NDEBUG
	// Write some info about the scenario tree
	for (int n = 0; n < scTree.nmbNodes; n++) {
		cout << "Node " << setw(2) << treeNodes[n]->nodeNmb << " has "
		     << "prob = " << setw(5) << treeNodes[n]->prob << " and "
		     << treeNodes[n]->children.size() << " children: ";
		for (int cn = 0; cn < (int) treeNodes[n]->children.size(); cn++)
			cout << setw(2) << treeNodes[n]->children[cn]->nodeNmb << " ";
		cout << endl;
	}
	#endif

	// create a "shortcut object" for the root
	assert (dynamic_cast<RootNode *>(treeNodes[0]) != NULL
					&& "The first node in treeNodes should be the Root");
	RootNode *ptRoot = dynamic_cast<RootNode *>(treeNodes[0]);

	ptRoot->make_objective_function();       // Create the objective func.
	model.setObjective(ptRoot->objFunction); // Set the objective
	model.attach();                          // Attach the model
	model->writeMps("investment");           // Write an MPS file

	// Solve the model
	model.minimize();

	// Report some results
	cout << endl << "Optimal objective value = " << model->getObjValue() << endl
			 << "First-stage x:";
	ptRoot->x.display();

	for (int n = 0; n < (int) treeNodes.size(); n++)
		delete treeNodes[n]; // cleaning

	return 0;
}
